# pylint: disable=all
import sys
from pathlib import Path
from importlib.metadata import metadata

import sphinx_rtd_theme


# -- Project information -----------------------------------------------------

package_metadata = metadata('drover')
project = package_metadata.get('name', '')
author = package_metadata.get('author', '')
version = package_metadata.get('version', '')
release = package_metadata.get('version', '')

if author:
    copyright = f'2020 {author} via Creative Commons Attribution 4.0 International License'
else:
    copyright = 'Creative Commons Attribution 4.0 International License'

epub_description = package_metadata.get('description', '')


# -- General configuration ---------------------------------------------------

default_role = 'py:obj'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.intersphinx',
    'sphinx_rtd_theme',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

intersphinx_mapping = {'python': ('https://docs.python.org/3', None)}


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

html_theme_path = ['_themes']

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


for path in templates_path + html_static_path:
    path = Path(path)
    path.mkdir(exist_ok=True)



import tempfile
from pathlib import Path

from invoke import task

_SOURCE_PATH = Path(__file__).parent.absolute()


@task
def deploy(context):
    with tempfile.TemporaryDirectory(prefix='.deploy-', dir=_SOURCE_PATH) as deploy_path:
        deploy_path = Path(deploy_path).absolute()
        context.run(
            f'docker run -v {_SOURCE_PATH}:/var/task -v {deploy_path}:/build '
            'lambci/lambda:build-python3.8 '
            'sh -c "set -x && '
            'python -m venv --clear /tmp/venv && '
            '/tmp/venv/bin/pip install wheel && '
            '/tmp/venv/bin/pip install --target /build -r requirements.txt ."')

        # The `drover` utility configuration may contain files relative to the source working directory.
        with context.cd(str(_SOURCE_PATH)):
            context.run(f'drover --install-path {deploy_path} production')


@task
def request(context):
    with tempfile.NamedTemporaryFile(prefix='output', dir=_SOURCE_PATH) as output_file:
        context.run(
            'aws --region us-east-1 lambda invoke '
            '--function-name "basic-lambda" --invocation-type "RequestResponse" '
            f'{output_file.name}')
        output_file.seek(0)
        print('Lambda Output:', output_file.read(), sep='\n')



# pylint: disable=protected-access
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch, MagicMock

import boto3
import pytest

from drover import Drover, SettingsError
from drover.models import Settings, Stage


def get_supported_compatible_runtime():
    return 'python3.8'


def get_basic_settings(expected_stage_name: str) -> Settings:
    expected_stage = Stage(
        region_name='region_name',
        function_name='function_name',
        compatible_runtime=get_supported_compatible_runtime(),
        function_file_patterns=[
            '^function.*'
        ])
    expected_settings = Settings(
        stages={
            expected_stage_name: expected_stage
        })
    return expected_settings


class TestDrover(TestCase):
    def test_init_with_valid_settings_and_invalid_stage_name(self):
        expected_invalid_stage_name = 'stage-invalid'
        expected_settings = get_basic_settings('stage')
        expected_compatible_runtime_library_path = Path('path')

        mock_boto3_client = MagicMock()

        with patch.object(Drover, '_get_runtime_library_path', return_value=expected_compatible_runtime_library_path), \
             patch.object(boto3, 'client', return_value=mock_boto3_client):
            with pytest.raises(SettingsError, match=r'^Invalid stage name.*'):
                Drover(expected_settings, expected_invalid_stage_name, interactive=False)

    def test_init_with_valid_settings_and_stage(self):
        expected_stage_name = 'stage'
        expected_settings = get_basic_settings(expected_stage_name)
        expected_stage = expected_settings.stages[expected_stage_name]
        expected_interactive = False
        expected_requirements_layer_name = 'function_name-requirements'
        expected_compatible_runtime_library_path = Path('path')

        mock_boto3_client = MagicMock()

        with patch.object(Drover, '_get_runtime_library_path', return_value=expected_compatible_runtime_library_path) as mock_get_runtime_library_path, \
             patch.object(boto3, 'client', return_value=mock_boto3_client) as mock_boto3_client_cls:
            drover = Drover(expected_settings, expected_stage_name, interactive=expected_interactive)
            mock_get_runtime_library_path.assert_called_once_with(expected_stage.compatible_runtime)
            mock_boto3_client_cls.assert_called_once_with('lambda', region_name=expected_stage.region_name)

        assert drover.settings == expected_settings
        assert drover.interactive == expected_interactive
        assert drover.stage == expected_stage
        assert drover.stage.requirements_layer_name == expected_requirements_layer_name
        assert drover.compatible_runtime_library_path == expected_compatible_runtime_library_path
        assert drover.lambda_client == mock_boto3_client

    def test_runtime_library_path_supports_python(self):
        for python_version in ('python3.6', 'python3.7', 'python3.8'):
            with self.subTest(python_version=python_version):
                assert Drover._get_runtime_library_path(python_version).name == 'python'



"""Models for settings and Amazon Web Services interactions"""
import re
from pathlib import Path
from typing import Mapping, Optional, Pattern, Sequence

from pydantic import BaseModel


class S3BucketFileVersion(BaseModel):
    bucket_name: str
    key: str
    version_id: Optional[str]


class S3BucketPath(BaseModel):
    region_name: str
    bucket_name: str
    prefix: str = ''


class Stage(BaseModel):
    region_name: str
    function_name: str
    compatible_runtime: str
    function_file_patterns: Sequence[Pattern]
    function_extra_paths: Sequence[Path] = []
    requirements_layer_name: Optional[str]
    supplemental_layer_arns: Sequence[str] = []
    package_exclude_patterns: Sequence[Pattern] = [re.compile(r'.*__pycache__.*')]
    upload_bucket: Optional[S3BucketPath]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if not self.requirements_layer_name:
            self.requirements_layer_name = f'{self.function_name}-requirements'


class Settings(BaseModel):
    stages: Mapping[str, Stage]



import ast
import codecs
import os.path

import setuptools


def read(*path_parts, iterate_lines=False):
    source_file_name = os.path.join(*path_parts)
    if not os.path.isfile(source_file_name):
        raise FileNotFoundError(source_file_name)
    with codecs.open(source_file_name, 'r') as source_file:
        return source_file.readlines() if iterate_lines else source_file.read()


def get_version(*path_parts) -> str:
    # See: <https://packaging.python.org/guides/single-sourcing-package-version/>
    for line in read(*path_parts, iterate_lines=True):
        if line.startswith('__version__'):
            return ast.parse(line).body[0].value.s
    raise RuntimeError('Unable to determine version.')


HERE = os.path.abspath(os.path.dirname(__file__))
VERSION = get_version(HERE, 'basic_lambda', '__init__.py')
LONG_DESCRIPTION = read(HERE, 'README.md')


setuptools.setup(
    name='basic_lambda',
    version=VERSION,
    author='Jeffrey Wilges',
    author_email='jeffrey@wilges.com',
    description='a basic Lambda that returns its version',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url='https://github.com/jwilges/drover',
    license='BSD',
    packages=setuptools.find_packages(exclude=['tests*']),
    python_requires='>=3.8',
    install_requires=[],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.8',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Networking',
    ],
)



VERSION = '0.7.2rc1'
NAME = 'drover'
AUTHOR = 'Jeffrey Wilges'
AUTHOR_EMAIL = 'jeffrey@wilges.com'
DESCRIPTION = 'drover is a command-line utility for deploying Python packages to Lambda functions.'
URL = 'https://github.com/jwilges/drover'
LICENSE = 'BSD'



import os
import zipfile
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch, MagicMock

from drover.io import ArchiveMapping, get_relative_file_names, write_archive


class TestGetRelativeFileNames(TestCase):
    def test_without_excludes(self):
        expected_source_path = '/'
        expected_relative_file_names = set((
            Path('file_0'),
            Path('path_a/file_a_0'),
            Path('path_a/file_a_1'),
            Path('path_b/file_b_0')))
        expected_walk = (
            ('/', ('path_a', 'path_b',), ('file_0',)),
            ('/path_a', (), ('file_a_0', 'file_a_1')),
            ('/path_b', (), ('file_b_0',)),
        )
        with patch.object(os, 'walk', return_value=expected_walk) as mock_walk:
            names = set(get_relative_file_names(expected_source_path))
            mock_walk.assert_called_once_with(str(expected_source_path))
            assert names == expected_relative_file_names


class TestWriteArchive(TestCase):
    def test_write_empty_archive(self):
        expected_archive_file_name = Path('archive.zip')
        mock_zip_file = MagicMock(spec=zipfile.ZipFile)
        with patch.object(zipfile, 'ZipFile') as mock_zip_file_cls:
            mock_zip_file_cls.return_value.__enter__.return_value = mock_zip_file
            write_archive(expected_archive_file_name, [])
            mock_zip_file_cls.assert_called_once_with(
                expected_archive_file_name, 'w',
                compression=zipfile.ZIP_DEFLATED,
                compresslevel=9)
            mock_zip_file.write.assert_not_called()

    def test_write_non_empty_archive(self):
        expected_archive_file_name = Path('archive.zip')
        expected_archive_mappings = [
            ArchiveMapping(source_file_name=Path('source/a'), archive_file_name=Path('archive/a')),
            ArchiveMapping(source_file_name=Path('source/b'), archive_file_name=Path('archive/b')),
        ]
        mock_zip_file = MagicMock(spec=zipfile.ZipFile)
        with patch.object(zipfile, 'ZipFile') as mock_zip_file_cls:
            mock_zip_file_cls.return_value.__enter__.return_value = mock_zip_file
            write_archive(expected_archive_file_name, expected_archive_mappings)
            mock_zip_file_cls.assert_called_once_with(
                expected_archive_file_name, 'w',
                compression=zipfile.ZIP_DEFLATED,
                compresslevel=9)
        for expected_archive_mapping in expected_archive_mappings:
            mock_zip_file.write.assert_any_call(
                filename=expected_archive_mapping.source_file_name,
                arcname=expected_archive_mapping.archive_file_name)



"""drover: a command-line utility to deploy Python packages to Lambda functions"""
import logging
import os
import re
import shutil
import sys
import tempfile
import zipfile
from http import HTTPStatus
from io import StringIO
from pathlib import Path
from typing import Any, BinaryIO, Callable, Iterable, List, Mapping, Optional, Pattern, Sequence

import boto3
import botocore
import tqdm
from pydantic import BaseModel

from drover.io import (
    ArchiveMapping,
    FunctionLayerMappings,
    format_file_size,
    get_digest,
    get_relative_file_names,
    write_archive,
)
from drover.models import S3BucketFileVersion, S3BucketPath, Settings, Stage

_logger = logging.getLogger(__name__)


class SettingsError(RuntimeError):
    """Base settings error"""


class UpdateError(RuntimeError):
    """Base update error"""


class Drover:
    """An interface to efficiently publish and update a Lambda function and requirements layer
    representation of a Python package directory"""
    def __init__(self, settings: Settings, stage: str, interactive: bool = False):
        self.settings = settings
        self.interactive = interactive

        if stage not in self.settings.stages:
            raise SettingsError(f'Invalid stage name: {stage}')

        self.stage = self.settings.stages[stage]
        self.compatible_runtime_library_path = Drover._get_runtime_library_path(self.stage.compatible_runtime)
        self.lambda_client = boto3.client('lambda', region_name=self.stage.region_name)

    def _get_function_layer_mappings(self, install_path: Path) -> FunctionLayerMappings:
        requirements_base_path = self.compatible_runtime_library_path
        function_file_patterns = self.stage.function_file_patterns

        requirements_mappings: List[ArchiveMapping] = []
        function_mappings: List[ArchiveMapping] = []
        for relative_file_name in get_relative_file_names(install_path, self.stage.package_exclude_patterns):
            source_file_name = install_path / relative_file_name
            if any([pattern.match(str(relative_file_name)) for pattern in function_file_patterns]):
                function_mappings.append(
                    ArchiveMapping(
                        source_file_name=source_file_name,
                        archive_file_name=relative_file_name))
            else:
                requirements_mappings.append(
                    ArchiveMapping(
                        source_file_name=source_file_name,
                        archive_file_name=requirements_base_path / relative_file_name))
        for extra_path in self.stage.function_extra_paths:
            for relative_file_name in get_relative_file_names(extra_path, self.stage.package_exclude_patterns):
                source_file_name = extra_path / relative_file_name
                function_mappings.append(
                    ArchiveMapping(
                        source_file_name=source_file_name,
                        archive_file_name=relative_file_name))

        requirements_digest = get_digest((mapping.source_file_name for mapping in requirements_mappings))
        function_digest = get_digest((mapping.source_file_name for mapping in function_mappings))

        if _logger.isEnabledFor(logging.DEBUG):
            def _log(header: str, mappings: Sequence[ArchiveMapping]):
                with StringIO() as message:
                    message.write(header)
                    for mapping in sorted(mappings, key=lambda item: item.archive_file_name):
                        message.write(f'  {mapping.archive_file_name}: {mapping.source_file_name}\n')
                    _logger.debug(message.getvalue())
            _log('Requirements file mappings:\n', requirements_mappings)
            _log('Function file mappings:\n', function_mappings)

        _logger.info('Requirements digest: %s', requirements_digest)
        _logger.info('Function digest: %s', function_digest)

        return FunctionLayerMappings(
            function_mappings=function_mappings,
            function_digest=function_digest,
            requirements_mappings=requirements_mappings,
            requirements_digest=requirements_digest)

    def update(self, install_path: Path) -> None:
        """Publish and/or update a Lambda function and/or requirements layer representation of a Python package directory

        Args:
            install_path: a Python package directory (e.g. via `pip install -t <install_path>`)"""

        if not install_path.is_dir():
            raise UpdateError(f'Install path is invalid: {install_path}')

        mappings = self._get_function_layer_mappings(install_path)

        try:
            function_response = self.lambda_client.get_function(FunctionName=self.stage.function_name)
        except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
            raise UpdateError(f'Unable to retrieve Lambda function "{self.stage.function_name}": {e}') from e

        function_arn = function_response['Configuration']['FunctionArn']
        function_layer_arns: List[str] = [layer['Arn'] for layer in function_response['Configuration'].get('Layers', [])]
        function_runtime = function_response['Configuration']['Runtime']
        function_tags: Mapping[str, str] = function_response.get('Tags', {})
        head_requirements_digest = function_tags.get('HeadRequirementsDigest')
        head_requirements_layer_arn = function_tags.get('HeadRequirementsLayerArn')
        head_function_digest = function_tags.get('HeadFunctionDigest')

        head_requirements_layer_arn_missing = True
        if head_requirements_layer_arn:
            try:
                self.lambda_client.get_layer_version_by_arn(Arn=head_requirements_layer_arn)
                head_requirements_layer_arn_missing = False
            except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
                _logger.warning('Unable to retrieve requirements layer "%s"; forcing re-upload.',
                                head_requirements_layer_arn)
                _logger.debug('', exc_info=e)

        should_upload_requirements = mappings.requirements_mappings and any((
            not head_requirements_digest,
            not head_requirements_layer_arn,
            head_requirements_layer_arn_missing,
            head_requirements_digest != mappings.requirements_digest))
        if should_upload_requirements:
            requirements_layer_arn = self._upload_requirements_archive(mappings.requirements_mappings,
                                                                       mappings.requirements_digest)
            function_tags['HeadRequirementsDigest'] = mappings.requirements_digest
            function_tags['HeadRequirementsLayerArn'] = requirements_layer_arn
        else:
            requirements_layer_arn = head_requirements_layer_arn
            function_tags.pop('HeadRequirementsDigest', None)
            function_tags.pop('HeadRequirementsLayerArn', None)
            _logger.info('Skipping requirements upload')

        head_function_layer_arns = [arn for arn in (*self.stage.supplemental_layer_arns,
                                                    head_requirements_layer_arn) if arn]

        if function_runtime != self.stage.compatible_runtime or function_layer_arns != head_function_layer_arns:
            _logger.info('Updating function resource...')
            try:
                self.lambda_client.update_function_configuration(
                    FunctionName=self.stage.function_name,
                    Runtime=self.stage.compatible_runtime,
                    Layers=head_function_layer_arns)
            except botocore.exceptions.BotoCoreError as e:
                raise UpdateError(f'Failed to update function "{self.stage.function_name}" runtime and layers: {e}') from e
            _logger.info('Updated function "%s" resource; runtime: "%s"; layers: %s',
                         self.stage.function_name, self.stage.compatible_runtime, head_function_layer_arns)

        if not head_function_digest or head_function_digest != mappings.function_digest:
            self._upload_function_archive(mappings.function_mappings)
            function_tags['HeadFunctionDigest'] = mappings.function_digest
        else:
            _logger.info('Skipping function upload')

        function_tags = {key: value for key, value in function_tags.items() if value}
        if function_tags:
            try:
                self.lambda_client.tag_resource(Resource=function_arn, Tags=function_tags)
            except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
                raise UpdateError(f'Unable to update tags for Lambda function "{self.stage.function_name}": {e}') from e

    def _upload_file_to_bucket(self, file_name: Path) -> S3BucketFileVersion:
        upload_bucket: S3BucketPath = self.stage.upload_bucket
        s3_client = boto3.client('s3', region_name=upload_bucket.region_name)
        file_size = float(file_name.stat().st_size)
        key = f'{upload_bucket.prefix}{file_name.name}'
        with tqdm.tqdm(total=file_size, unit='B', unit_divisor=1024, unit_scale=True, leave=True,
                       disable=not self.interactive) as progress:
            s3_client.upload_file(
                Filename=str(file_name),
                Bucket=upload_bucket.bucket_name,
                Key=key,
                Callback=progress.update)

        response = s3_client.head_object(
            Bucket=upload_bucket.bucket_name,
            Key=key
        )
        return S3BucketFileVersion(
            bucket_name=upload_bucket.bucket_name,
            key=key,
            version_id=response.get('VersionId'))

    def _delete_file_from_bucket(self, bucket_file: S3BucketFileVersion):
        upload_bucket = self.stage.upload_bucket
        s3_client = boto3.client('s3', region_name=upload_bucket.region_name)
        arguments = {
            'Bucket': bucket_file.bucket_name,
            'Key': bucket_file.key,
        }
        if bucket_file.version_id:
            arguments['VersionId'] = bucket_file.version_id
        s3_client.delete_object(**arguments)

    def _upload_requirements_archive(self, archive_mappings: Sequence[ArchiveMapping], archive_digest: str) -> str:
        archive_handle, archive_file_name = tempfile.mkstemp(suffix='.zip', prefix='_requirements-')
        archive_file_name = Path(archive_file_name)

        def _upload() -> str:
            try:
                write_archive(archive_file_name, archive_mappings)
            finally:
                os.close(archive_handle)

            if self.stage.upload_bucket:
                _logger.info('Uploading requirements layer archive...')
                try:
                    bucket_file = self._upload_file_to_bucket(archive_file_name)
                    file_arguments = {
                        'S3Bucket': bucket_file.bucket_name,
                        'S3Key': bucket_file.key,
                    }
                    if bucket_file.version_id:
                        file_arguments['S3ObjectVersion'] = bucket_file.version_id
                except (botocore.exceptions.ClientError, boto3.exceptions.S3UploadFailedError) as e:
                    _logger.error('Failed to upload requirements archive to bucket; falling back to direct file upload.')
                    _logger.debug('', exc_info=e)
                    bucket_file = None

            if not bucket_file:
                with open(archive_file_name, 'rb') as archive_file:
                    file_arguments = {'ZipFile': archive_file.read()}

            archive_description = f'Requirements layer for {self.stage.function_name}; digest: {archive_digest}'
            _logger.info('Publishing requirements layer...')
            try:
                response = self.lambda_client.publish_layer_version(
                    LayerName=self.stage.requirements_layer_name,
                    Description=archive_description,
                    Content=file_arguments,
                    CompatibleRuntimes=[self.stage.compatible_runtime])
            except botocore.exceptions.BotoCoreError as e:
                raise UpdateError(f'Failed to publish requirements layer for {self.stage.function_name}: {e}') from e
            finally:
                if bucket_file:
                    self._delete_file_from_bucket(bucket_file)

            layer_version_arn = response['LayerVersionArn']
            layer_size_text = format_file_size(float(response['Content']['CodeSize']))
            _logger.info('Published requirements layer "%s"; size: %s; ARN: %s',
                         self.stage.requirements_layer_name, layer_size_text, layer_version_arn)

            return layer_version_arn

        try:
            return _upload()
        finally:
            archive_file_name.unlink()

    def _upload_function_archive(self, archive_mappings: Sequence[ArchiveMapping]) -> str:
        archive_handle, archive_file_name = tempfile.mkstemp(suffix='.zip', prefix='_function-')
        archive_file_name = Path(archive_file_name)

        def _upload() -> str:
            try:
                write_archive(archive_file_name, archive_mappings)
            finally:
                os.close(archive_handle)

            if self.stage.upload_bucket:
                _logger.info('Uploading function archive...')
                try:
                    bucket_file = self._upload_file_to_bucket(archive_file_name)
                    file_arguments = {
                        'S3Bucket': bucket_file.bucket_name,
                        'S3Key': bucket_file.key,
                    }
                    if bucket_file.version_id:
                        file_arguments['S3ObjectVersion'] = bucket_file.version_id
                except (botocore.exceptions.ClientError, boto3.exceptions.S3UploadFailedError) as e:
                    _logger.error('Failed to upload function archive to bucket; falling back to direct file upload.')
                    _logger.debug('', exc_info=e)
                    bucket_file = None

            if not bucket_file:
                with open(archive_file_name, 'rb') as archive_file:
                    file_arguments = {'ZipFile': archive_file.read()}

            _logger.info('Updating function resource...')
            try:
                response = self.lambda_client.update_function_code(
                    FunctionName=self.stage.function_name,
                    Publish=False,
                    DryRun=False,
                    **file_arguments)
            except botocore.exceptions.BotoCoreError as e:
                raise RuntimeError(f'Failed to update function code for {self.stage.function_name}: {e}') from e
            finally:
                if bucket_file:
                    self._delete_file_from_bucket(bucket_file)

            function_arn = response['FunctionArn']
            function_size_text = format_file_size(float(response['CodeSize']))
            _logger.info('Updated function "%s" resource; size: %s; ARN: %s',
                         self.stage.function_name, function_size_text, function_arn)

            return function_arn

        try:
            return _upload()
        finally:
            archive_file_name.unlink()

    @staticmethod
    def _get_runtime_library_path(runtime: str) -> Path:
        python_pattern = re.compile(r'^python\d+\.\d+$')
        if python_pattern.match(runtime):
            return Path('python')
        raise NotImplementedError(f'Unsupported runtime: {runtime}')



import shutil
import sys
from pathlib import Path
from typing import Sequence

from invoke import call, task

_SOURCE_PATH = 'drover'
_PROXY_SEPARATOR = '--'

def add_proxy_arguments(argument: str) -> Sequence[str]:
    if _PROXY_SEPARATOR in sys.argv:
        return f'{argument} {" ".join(sys.argv[sys.argv.index(_PROXY_SEPARATOR) + 1:])}'
    return argument


@task
def bandit(context):
    excluded_paths = ','.join(('*/tests/*', ))
    context.run(add_proxy_arguments(f'bandit --exclude {excluded_paths} --recursive {_SOURCE_PATH}'))


@task
def yapf(context, check = False):
    arguments = '--recursive'
    arguments += ' --diff' if check else ' --in-place'
    context.run(add_proxy_arguments(f'yapf {arguments} {_SOURCE_PATH}'), hide='stdout' if check else None)


@task
def isort(context, check = False):
    arguments = ' --check --diff' if check else ''
    context.run(add_proxy_arguments(f'isort {arguments} --src {_SOURCE_PATH} {_SOURCE_PATH}'))


@task
def test(context):
    context.run(add_proxy_arguments('pytest'))


@task
def lint(context):
    context.run(add_proxy_arguments(f'pylint --disable C,R,fixme {_SOURCE_PATH}'))


@task
def mypy(context):
    context.run(add_proxy_arguments(f'mypy {_SOURCE_PATH}'))


@task(help={'upgrade': 'try to upgrade all dependencies to their latest versions'})
def compile_requirements(context, upgrade = False):
    """Compile requirements.txt and requirements.dev.txt from their .in specifications"""
    arguments = '-U' if upgrade else ''
    context.run(add_proxy_arguments(f'pip-compile {arguments}'))
    context.run(add_proxy_arguments(f'pip-compile {arguments} requirements.dev.in'))


def coverage_base(context, mode, run):
    if run:
        context.run(add_proxy_arguments('coverage run -m pytest'))
    context.run(f'coverage {mode}')


@task
def coverage(context, run = True):
    coverage_base(context, 'report', run)


@task
def coverage_xml(context, run = True):
    coverage_base(context, 'xml', run)


@task
def coverage_html(context, run = True):
    coverage_base(context, 'html', run)


@task(call(isort, check=True), call(yapf, check=True), lint, mypy, bandit, coverage)
def ci(context): ...


@task
def docs_html(context):
    source_path = (Path(__file__).parent / 'docs').absolute()
    context.run(f'sphinx-apidoc --force --output-dir "{str(source_path)}" "{_SOURCE_PATH}"')
    context.run('python setup.py build_sphinx')


@task
def wheel(context):
    context.run('pip install -c requirements.dev.txt wheel')
    release_paths = [Path(path) for path in ('build', 'dist')]
    for path in release_paths:
        if path.exists():
            shutil.rmtree(path)
    context.run('python setup.py sdist bdist_wheel')



"""Command-line interface functionality for the Drover interface"""
import argparse
import logging
import sys
from pathlib import Path

import yaml
from pydantic import ValidationError

from drover import Drover, SettingsError, UpdateError
from drover.__metadata__ import VERSION
from drover.models import Settings

_logger = logging.getLogger(__name__)


def _parse_arguments():
    parser = argparse.ArgumentParser(description=__doc__.partition('\n')[0])
    parser.add_argument('--version', '-V', action='version', version=f'%(prog)s {VERSION}')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--verbose', '-v', action='count', default=0, help='increase output verbosity')
    group.add_argument('--quiet', action='store_true', help='disable output')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--interactive', action='store_true', help='enable interactive output (i.e. for a PTY)')
    group.add_argument('--non-interactive', action='store_true', help='disable interactive output')

    parser.add_argument('--settings-file', default=Path('drover.yml'), type=Path,
                        help='Settings file name (default: "drover.yml")')
    parser.add_argument('--install-path', default=Path(), type=Path,
                        help='Package install path (e.g. from "pip install -t"; default: working directory)')
    parser.add_argument('stage', type=str)
    return parser.parse_args()


def _parse_settings(settings_file_name: Path) -> Settings:
    try:
        with open(settings_file_name, 'r') as settings_file:
            return Settings.parse_obj(yaml.safe_load(settings_file))
    except (ValueError, ValidationError) as e:
        _logger.error('Settings file is invalid: %s', e)
        _logger.debug('', exc_info=e)
        sys.exit(1)
    except FileNotFoundError as e:
        _logger.error('Settings file does not exist: %s', e)
        _logger.debug('', exc_info=e)
        sys.exit(1)


def main():
    """The main command-line entry point for the Drover interface"""
    arguments = _parse_arguments()

    if not arguments.quiet:
        logging.basicConfig(format='%(message)s', stream=sys.stdout)
        logging_level = max(1, logging.INFO - (10 * arguments.verbose))
        logging.getLogger(__name__.split('.')[0]).setLevel(logging_level)

    interactive = True if arguments.interactive else False if arguments.non_interactive else sys.__stdin__.isatty()

    settings_file_name = arguments.settings_file
    install_path: Path = arguments.install_path

    settings: Settings = _parse_settings(settings_file_name)

    try:
        drover = Drover(settings, arguments.stage, interactive=interactive)
        drover.update(install_path)
    except SettingsError as e:
        _logger.error('Initialization failed: %s', e)
        _logger.debug('', exc_info=e)
        sys.exit(1)
    except UpdateError as e:
        _logger.error('Update failed: %s', e)
        _logger.debug('', exc_info=e)
        sys.exit(1)



"""Conditional patches for Python's `copy` module"""
import copy
import re
import sys

if sys.version_info < (3, 7):
    # See: <https://stackoverflow.com/questions/6279305/typeerror-cannot-deepcopy-this-pattern-object>
    copy._deepcopy_dispatch[type(re.compile(''))] = lambda r, _: r # pylint: disable=protected-access



# pylint: disable=protected-access
from dataclasses import dataclass
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

import drover
import drover.cli

from tests.test_drover import get_basic_settings


@dataclass
class MockArguments:
    stage: str
    quiet: bool
    interactive: bool = True
    non_interactive: bool = False
    settings_file: Path = Path('drover.yml')
    install_path: Path = Path('.')


class TestDroverCLI(TestCase):
    def test_basic_run(self):
        expected_stage_name = 'stage'
        expected_settings = get_basic_settings(expected_stage_name)
        expected_interactive = True
        mock_arguments = MockArguments(
            stage='stage',
            quiet=True,
            interactive=expected_interactive, non_interactive=not expected_interactive)
        with patch.object(drover.cli, '_parse_arguments', return_value=mock_arguments) as mock_parse_arguments, \
             patch.object(drover.cli, '_parse_settings', return_value=expected_settings) as mock_parse_settings, \
             patch.object(drover.cli, 'Drover', autospec=drover.Drover) as mock_drover:
            drover.cli.main()
            mock_drover.assert_called_with(expected_settings, expected_stage_name, interactive=expected_interactive)
            mock_parse_arguments.assert_called()
            mock_parse_settings.assert_called_with(mock_arguments.settings_file)



import codecs
import re
import sys
from pathlib import Path
from typing import List, MutableMapping, Optional, Sequence, Tuple, Type

import setuptools


def _read(source_file_name: Path):
    if not source_file_name.is_file():
        raise FileNotFoundError(source_file_name)
    with codecs.open(str(source_file_name), 'r') as source_file:
        return source_file.read()


HERE = Path().parent.absolute()
PACKAGE_PATH = HERE / 'drover'

LONG_DESCRIPTION = _read(HERE / 'README.md')
METADATA = {
    'name': 'drover',
    'author': 'Jeffrey Wilges',
    'author_email': 'jeffrey@wilges.com',
    'description': 'drover is a command-line utility for deploying Python packages to Lambda functions.',
    'url': 'https://github.com/jwilges/drover',
    'license': 'BSD'
}
METADATA_TEMPLATE = '\n'.join((
    "VERSION = '{version}'",
    *(
        f"{key.upper()} = '{value.replace('{', '{{').replace('}', '}}')}'"
        for key, value in METADATA.items()
    ),
    '' # final newline
))

OPTIONAL_COMMAND_CLASSES: MutableMapping[str, Type] = {}
OPTIONAL_COMMAND_OPTIONS: MutableMapping[str, MutableMapping[str, Sequence[str]]] = {}

try:
    from sphinx.setup_command import BuildDoc
    OPTIONAL_COMMAND_CLASSES['build_sphinx'] = BuildDoc
    OPTIONAL_COMMAND_OPTIONS['build_sphinx'] = {
        'source_dir': ('setup.py', 'docs'),
        'build_dir': ('setup.py', 'docs/_build')
    }
except ImportError:
    pass


class ValidateTagCommand(setuptools.Command):
    """A validator that ensures the package version both is in the canonical form per
    PEP-440 and matches the current git tag"""
    description = 'validate that the package version matches the current git tag'
    user_options: List[Tuple[str, Optional[str], str]] = [
        ('output-azure-variables', None, 'Output Azure Pipeline version variables'),
    ]

    VERSION_PATTERN = r'''
        v?
        (?:
            (?:(?P<epoch>[0-9]+)!)?                           # epoch
            (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
            (?P<pre>                                          # pre-release
                [-_\.]?
                (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
                [-_\.]?
                (?P<pre_n>[0-9]+)?
            )?
            (?P<post>                                         # post release
                (?:-(?P<post_n1>[0-9]+))
                |
                (?:
                    [-_\.]?
                    (?P<post_l>post|rev|r)
                    [-_\.]?
                    (?P<post_n2>[0-9]+)?
                )
            )?
            (?P<dev>                                          # dev release
                [-_\.]?
                (?P<dev_l>dev)
                [-_\.]?
                (?P<dev_n>[0-9]+)?
            )?
        )
        (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
    '''
    VERSION_FORMAT = re.compile(
        r'^\s*' + VERSION_PATTERN + r'\s*$',
        re.VERBOSE | re.IGNORECASE,
    )

    def initialize_options(self):
        self.output_azure_variables = False

    def finalize_options(self):
        pass

    def run(self):
        """Warn and exit if the package version either:
            a) is not in the canonical format per PEP-440 or,
            b) does not match any HEAD git tag version."""
        import setuptools_scm
        version = setuptools_scm.get_version(relative_to=Path(__file__))

        if self.output_azure_variables:
            print(f'##vso[task.setvariable variable=is_prerelease;isOutput=true;]{self.is_prerelease(version)!s}')

        if not self.is_canonical(version):
            self.warn(f'package version ({version}) is not in the canonical format per PEP-440')
            sys.exit(1)

    @classmethod
    def is_prerelease(cls, version: str) -> bool:
        """Return true if `version` is a 'pre', 'dev', or 'local' release per PEP-440."""
        version_match = cls.VERSION_FORMAT.match(version)
        return (
            version_match is None or
            any(
                value for key, value in version_match.groupdict().items()
                if key in ('pre', 'dev', 'local')
            )
        )

    @staticmethod
    def is_canonical(version: str) -> bool:
        """Return true if `version` is canonical per PEP-440."""
        return re.match(r'^([1-9][0-9]*!)?(0|[1-9][0-9]*)(\.(0|[1-9][0-9]*))*((a|b|rc)(0|[1-9][0-9]*))?(\.post(0|[1-9][0-9]*))?(\.dev(0|[1-9][0-9]*))?$', version) is not None


setuptools.setup(
    **METADATA,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(exclude=['tests*']),
    entry_points={
        'console_scripts': ['drover=drover.cli:main'],
    },
    use_scm_version={
        'relative_to': Path(__file__),
        'write_to': PACKAGE_PATH / '__metadata__.py',
        'write_to_template': METADATA_TEMPLATE,
    },
    setup_requires=['setuptools_scm'],
    python_requires='>=3.6',
    install_requires=[
        'boto3>=1.12',
        'botocore>=1.15',
        'pydantic>=1.4',
        'pyyaml>=5.3',
        'tqdm>=4.44',
        'dataclasses; python_version < "3.7"'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Build Tools',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Networking',
        'Topic :: System :: Software Distribution',
        'Topic :: Utilities'
    ],
    cmdclass={
        'validate_tag': ValidateTagCommand,
        **OPTIONAL_COMMAND_CLASSES
    },
    command_options={
        **OPTIONAL_COMMAND_OPTIONS
    },
)



import nox

nox.options.sessions = ['dev-3.8']
nox.options.error_on_missing_interpreters = True


@nox.session(python=['3.6', '3.7', '3.8'])
def dev(session):
    session.install('-r', 'requirements.dev.txt')
    session.install('-c', 'requirements.txt', '-e', '.')
    if not session.posargs:
        return
    session.run('invoke', *session.posargs)



"""Generic functionality related to files and I/O"""
import csv
import email.parser
import email.policy
import hashlib
import os
import re
import zipfile
from dataclasses import dataclass, field
from pathlib import Path
from typing import Iterable, Pattern, Sequence


@dataclass
class ArchiveMapping:
    """A mapping between an archive file name and its corresponding source filesystem path"""
    source_file_name: Path
    archive_file_name: Path


@dataclass
class FunctionLayerMappings:
    """A function and requirements layer mapping and digest container"""
    function_mappings: Sequence[ArchiveMapping] = field(default=list)
    function_digest: str = None
    requirements_mappings: Sequence[ArchiveMapping] = field(default=list)
    requirements_digest: str = None


def format_file_size(size_in_bytes: float) -> str:
    """Return a string representation of the specified size as its largest 2^10 representation

    Examples:
        >>> format_file_size(2048)
        '2.00 KiB'
        >>> format_file_size(16252928.0)
        '15.50 MiB'

    Args:
        size_in_bytes: a size in bytes

    Returns: a string representation of the specified size as its largest 2^10 representation"""
    for unit in ('B', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB'):
        if abs(size_in_bytes) < 1024.0:
            return f'{size_in_bytes:.2f} {unit}'
        size_in_bytes /= 1024.0
    return f'{size_in_bytes:.2f} YiB'


def get_digest(source_file_names: Sequence[Path], block_size: int = 8192) -> str:
    """Return a SHA256 hash composed from the content of all source files.

    Args:
        source_file_names: A sequence of source file paths

    Returns: A SHA256 hash composed from the content of all source files."""
    # See the PEP-376 RECORD file specification: <https://www.python.org/dev/peps/pep-0376/#record>
    package_record_pattern = re.compile(r'\.dist-info/RECORD$')
    egg_information_pattern = re.compile(r'\.egg-info/PKG-INFO$')
    digest = hashlib.sha256()
    full = set(source_file_names)
    done = set()
    if not full:
        return None
    for source_file_name in sorted(full):
        if package_record_pattern.search(str(source_file_name)):
            package_parent_path = source_file_name.parent.parent
            with open(source_file_name, 'r', buffering=block_size) as record:
                reader = csv.reader(record, delimiter=',', quotechar='"', lineterminator=os.linesep)
                for item in reader:
                    item_name, item_hash, _other = item[:3]
                    source_file_name = package_parent_path / item_name
                    if item_hash and source_file_name in full:
                        digest.update((str(item_name) + item_hash).encode())
                        done.add(source_file_name)
    remaining = full - done
    for source_file_name in sorted(remaining):
        with open(source_file_name, 'rb', buffering=block_size) as source_file:
            if egg_information_pattern.search(str(source_file_name)):
                # Ensure deterministic field order from PKG-INFO files
                # See: https://www.python.org/dev/peps/pep-0314/#including-metadata-in-packages
                parser = email.parser.BytesHeaderParser(policy=email.policy.default)
                source_headers = sorted(parser.parse(source_file).items())
                for header, value in source_headers:
                    digest.update(header.encode())
                    digest.update(value.encode())
            else:
                digest.update(source_file.read())

    return digest.hexdigest()


def get_relative_file_names(source_path: Path, exclude_patterns: Sequence[Pattern] = None) -> Iterable[Path]:
    """Return an unsorted iterable of files recursively beneath the source path

    Args:
        source_path: a filesystem path from which to recursively iterate all files
        exclude_patterns: an optional sequence of regular expressions which will be used to exclude files

    Returns: an unsorted iterable of files recursively beneath the source path"""
    exclude_patterns = exclude_patterns or []
    for root, _directory_names, file_names in os.walk(source_path):
        for file_name in file_names:
            relative_file_name = Path(os.path.join(root, file_name)).relative_to(source_path)
            if not any([pattern.match(str(relative_file_name)) for pattern in exclude_patterns]):
                yield relative_file_name


def write_archive(archive_file_name: Path, archive_mappings: Iterable[ArchiveMapping]) -> None:
    """Write a zip file archive composed of the specified archive file mappings

    Args:
        archive_file_name: a writable file
        archive_mappings: an iterable of mappings of filesystem file names to archive file names"""
    with zipfile.ZipFile(archive_file_name, 'w', compression=zipfile.ZIP_DEFLATED, compresslevel=9) as archive:
        for mapping in archive_mappings:
            archive.write(filename=mapping.source_file_name, arcname=mapping.archive_file_name)
